# -*- coding: utf-8 -*-

import datetime
import pathlib
import random
import re
import shutil

import bs4
import requests
from jinja2 import Environment, FileSystemLoader
from zimscraperlib.image.convertion import convert_image
from zimscraperlib.image.transformation import resize_image
from zimscraperlib.inputs import handle_user_provided_file
from zimscraperlib.zim.items import URLItem

from .constants import DEFAULT_HOMEPAGE, MAX_HTTP_404_THRESHOLD, ROOT_DIR, Conf
from .shared import Global, GlobalMixin, logger
from .utils import (
    article_ident_for,
    cat_ident_for,
    fix_pagination_links,
    get_categorylisting_url,
    get_digest,
    get_footer_crumbs_from,
    get_footer_links_from,
    get_soup,
    get_soup_of,
    get_subcategories_from,
    normalize_ident,
    parse_css,
    setup_s3_and_check_credentials,
    soup_link_finder,
    to_url,
)


class DomIntegrityError(Exception):
    pass


class wikihow2zim(GlobalMixin):
    def __init__(self, **kwargs):

        Global.conf = Conf(**kwargs)
        for option in Global.conf.required:
            if getattr(Global.conf, option) is None:
                raise ValueError(f"Missing parameter `{option}`")

        # jinja2 environment setup
        self.env = Environment(
            loader=FileSystemLoader(ROOT_DIR.joinpath("templates")), autoescape=True
        )
        self.env.filters["digest"] = get_digest

        # jinja context that we'll pass to all templates
        self.env_context = {"conf": Global.conf}

        # set of all categories we've seen link for
        # used to prevent processing a category twice as they cross-link each other
        self.categories = set()
        # idem, articles can be linked from different categories. we need
        # to keep track of which we processed
        self.articles = set()
        # idem, resources from CSS that we'll downloaded.
        # Source HTML references a dynamic CSS that is built using a varietyof features
        # so it's very common different CSS urls references the same resources (imgs)
        self.resources_digests = set()
        # idem, list of URLs which returned HTTP 404.
        # There are legit scenarios for 404 on wikiHow: login pages
        # we need to track them for later use
        self.missing_articles = set()
        self.missing_categories = set()
        # set of all articles we've seen in related-articles links.
        # we'll go over this at end of categories scrape.
        # those left are not listed in any category
        self.related_articles = set()

    @property
    def build_dir(self):
        return self.conf.build_dir

    def cleanup(self):
        """Remove temp files and release resources before exiting"""
        if not self.conf.keep_build_dir:
            logger.debug(f"Removing {self.build_dir}")
            shutil.rmtree(self.build_dir, ignore_errors=True)

    def get_style_urls(self, soup) -> object:
        """paths list of found CSS link content+resources appended to ZIM"""
        list_css = [
            to_url(lnk.attrs["href"]) for lnk in soup.find_all(soup_link_finder)
        ]
        for url in list_css:
            self.add_css(url)
        return list_css

    def get_online_metadata(self):
        """metadata from online website, looking at homepage source code"""
        logger.debug("Fecthing website metdata")

        soup, _ = get_soup("/")

        # external (<link />) and inline (<style/>) CSS resources
        linked_styles = []
        for link in soup.find_all(soup_link_finder):
            linked_styles.append(to_url(link.attrs["href"]))

        inline_styles = ""
        for style in soup.find_all("style", src=False):
            inline_styles += "\n" + style.string

        return {
            "dir": soup.find("html").attrs["dir"],
            "category_prefix": normalize_ident(
                soup.select("#hp_categories_list a")[0].attrs.get("href")
            ).split(":", 1)[0][1:],
            "homepage_name": normalize_ident(
                soup.select("a#header_logo")[0].attrs.get("href")
            ).replace("/", ""),
            "title": soup.find("title").string,
            "description": soup.find("meta", attrs={"name": "description"}).attrs.get(
                "content"
            ),
            "icon": to_url(soup.find("link", rel="apple-touch-icon").attrs.get("href")),
            "favicon": to_url(soup.find("link", rel="shortcut icon").attrs.get("href")),
            "logo": to_url(soup.select("a#footer_logo img")[0].attrs["src"]),
            "inline_styles": inline_styles,
            "linked_styles": linked_styles,
            "url_special_category": get_categorylisting_url(),
            "footer_links": get_footer_links_from(soup),
        }

    def sanitize_inputs(self):
        """input & metadata sanitation"""
        logger.debug("Checking user-provided metadata")

        if not self.conf.name:
            self.conf.name = "wikihow_{lang}_{selection}".format(
                lang=self.conf.language["iso-639-1"],
                selection="selection" if self.conf.categories else "all",
            )

        period = datetime.datetime.now().strftime("%Y-%m")
        if self.conf.fname:
            # make sure we were given a filename and not a path
            self.conf.fname = pathlib.Path(self.conf.fname.format(period=period))
            if pathlib.Path(self.conf.fname.name) != self.conf.fname:
                raise ValueError(f"filename is not a filename: {self.conf.fname}")
        else:
            self.conf.fname = f"{self.conf.name}_{period}.zim"

        if not self.conf.title:
            self.conf.title = self.metadata["title"]
        self.conf.title = self.conf.title.strip()

        if not self.conf.description:
            self.conf.description = self.metadata["description"]
        self.conf.description = self.conf.description.strip()

        if not self.conf.author:
            self.conf.author = "wikiHow"
        self.conf.author = self.conf.author.strip()

        if not self.conf.publisher:
            self.conf.publisher = "openZIM"
        self.conf.publisher = self.conf.publisher.strip()

        self.conf.tags = list(
            set(
                self.conf.tag
                + ["_category:wikihow", "wikihow", "_videos:yes", "_pictures:yes"]
            )
        )

    def add_illustrations(self):
        logger.debug("Adding illustrations")

        src_illus_fpath = self.build_dir / "illustration"

        # if user provided a custom favicon, retrieve that
        if not self.conf.icon:
            self.conf.icon = self.metadata["icon"]
        handle_user_provided_file(source=self.conf.icon, dest=src_illus_fpath)

        # convert to PNG (might already be PNG but it's OK)
        illus_fpath = src_illus_fpath.with_suffix(".png")
        convert_image(src_illus_fpath, illus_fpath)

        # resize to appropriate size (ZIM uses 48x48 so we double for retina)
        for size in (96, 48):
            resize_image(illus_fpath, width=size, height=size, method="thumbnail")
            with open(illus_fpath, "rb") as fh:
                with self.lock:
                    self.creator.add_illustration(size, fh.read())

        # download and add actual favicon (ICO file)
        favicon_fpath = self.build_dir / "favicon.ico"
        handle_user_provided_file(source=self.metadata["favicon"], dest=favicon_fpath)
        with self.lock:
            self.creator.add_item_for("favicon.ico", fpath=favicon_fpath)

        # download apple-touch-icon
        with self.lock:
            self.creator.add_item(
                URLItem(url=self.metadata["icon"], path="apple-touch-icon.png")
            )

    def get_from_cache(self, url: str) -> bytes:
        """retrieve from local cache if present, otherwise download it first

        Only useful for devel/debug because there are many resources linked
        to assets and source website is quite slow"""
        fpath = self.build_dir.joinpath(f"cache_{get_digest(url)}")
        if fpath.exists():
            with open(fpath, "rb") as fh:
                return fh.read()

        with open(fpath, "wb") as fh:
            content = requests.get(url).content
            fh.write(content)
            return content

    def add_css(self, url: str, inline: bool = False) -> str:
        """Download and add a CSS URL/text, including all its dependencies

        All url() resources from the passed CSS URL/string will be fetched
        and added to ZIM as well. url() will be rewritten in the CSS source
        before adding to Zim"""

        if not inline:
            url = to_url(url)

        # skip if we already added it. Can be referenced from multiple pages
        digest = get_digest(url)
        if digest in self.resources_digests:
            return

        # fetch and transform source CSS
        source = url if inline else self.get_from_cache(url).decode("UTF-8")
        content, resources = parse_css(source)

        # fetch and add to Zim all its resources
        for rsc_url, rsc_path in set(resources):
            rsc_url = to_url(rsc_url)
            rsc_digest = get_digest(rsc_url)

            # skip resource if already handled
            if rsc_digest in self.resources_digests:
                continue

            try:
                with self.lock:
                    # specifically don't specify mimetype here so that scraperlib
                    # determines it from content. We may encounter PNG and SVG
                    self.creator.add_item_for(
                        path=rsc_path,
                        content=self.get_from_cache(rsc_url),
                    )
            except Exception:
                pass  # many are just not working at all
            else:
                self.resources_digests.add(rsc_digest)
                logger.debug(f"> {rsc_path}")

        path = f"assets/{digest}.css"
        with self.lock:
            self.creator.add_item_for(
                path=path,
                content=content,
                mimetype="text/css",
            )
        logger.debug(f"> {path}")
        self.resources_digests.add(digest)
        return digest

    def add_assets(self):
        """download and add site-wide assets, identified in metadata step"""
        logger.info("Adding assets")

        with self.lock:
            self.creator.add_item(
                URLItem(path="assets/logo", url=self.metadata["logo"])
            )
            # external link icons
            for direction in ("ltr", "rtl"):
                url = (
                    f"https://en.wikipedia.org/w/skins/Vector/resources/common/images/"
                    f"external-link-{direction}-icon.svg"
                )
                self.creator.add_item(
                    URLItem(path=f"assets/external-link-{direction}-icon.svg", url=url)
                )

        # handle the external and inline CSS found in homepage
        self.metadata["inline_digest"] = self.add_css(
            self.metadata["inline_styles"], inline=True
        )
        for url in self.metadata["linked_styles"]:
            self.add_css(url)

        # Articles have a custom inline CSS
        soup, _ = get_soup("/wikihow:About-wikiHow")
        self.metadata["article_inline_digest"] = self.add_css(
            "\n".join([style.string for style in soup.find_all("style", src=False)]),
            inline=True,
        )

        # recursively add our own assets, at a path identical to position in repo
        assets_root = pathlib.Path(ROOT_DIR.joinpath("assets"))
        for fpath in assets_root.glob("**/*"):
            if not fpath.is_file():
                continue
            path = str(fpath.relative_to(ROOT_DIR))

            logger.debug(f"> {path}")
            with self.lock:
                self.creator.add_item_for(path=path, fpath=fpath)

    def build_categories_list(self):
        """Parses Sitemap online to get a list of root-level categories"""
        logger.info("Building list of root categories")

        soup, _ = get_soup("/Special:CategoryListing")
        self.conf.categories = [
            cat_ident_for(link.attrs["href"]) for link in soup.select("#catlist a")
        ]

    def build_exclude_lists(self):
        """Using provided path/URL from --excldude, build list of exclusion

        List is stored on rewriter as it is responsible for removing links to
        those articles and categories.
        File must contain Article ID (URL path) or Category: category ID"""
        if not self.conf.exclude:
            return

        logger.info(f"Building exclusion list from {self.conf.exclude}")
        excludes_fpath = self.build_dir / "excludes.lst"
        handle_user_provided_file(source=self.conf.exclude, dest=excludes_fpath)

        with open(excludes_fpath, "r") as fh:
            for line in fh.readlines():
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                if line.startswith("Category:"):
                    self.rewriter.excluded_categories.add(line.split("Category:", 1)[1])
                else:
                    self.rewriter.excluded_articles.add(line)

        logger.info(
            f"> {len(self.rewriter.excluded_articles)} articles and "
            f"{len(self.rewriter.excluded_categories)} categories excluded."
        )

    def add_homepage(self):
        logger.info("Building homepage")

        soup, _ = get_soup("/Special:CategoryListing")

        # CategoryListing has custom CSS
        linked_styles = [
            to_url(lnk.attrs["href"]) for lnk in soup.find_all(soup_link_finder)
        ]
        for url in linked_styles:
            self.add_css(url)

        # extract and clean main content
        content = soup.select("div#content_wrapper")[0]
        _ = [script.decompose() for script in content.find_all("script")]

        page = self.env.get_template("home.html").render(
            to_root="./",
            body_classes=" ".join(soup.find("body").attrs.get("class", [])),
            content=self.rewriter.rewrite(content.decode_contents(), to_root="./"),
            page_linked_styles=linked_styles,
            viewport_classes=" ".join(
                soup.find(attrs={"id": "mw-mf-viewport"}).attrs.get("class", [])
            ),
            footer_links=self.metadata["footer_links"],
            title=self.conf.title,
            **self.env_context,
        )

        with self.lock:
            self.creator.add_item_for(
                path=self.metadata["homepage_name"],
                title=self.conf.title,
                content=page,
                mimetype="text/html",
                is_front=True,
            )
            self.creator.add_redirect(
                path=self.metadata["url_special_category"], target_path=DEFAULT_HOMEPAGE
            )

            if DEFAULT_HOMEPAGE != self.metadata["homepage_name"]:
                self.creator.add_redirect(
                    path=DEFAULT_HOMEPAGE, target_path=self.metadata["homepage_name"]
                )

    def scrape_footer_articles(self):
        """Scrape and create all pages found in footer links"""
        for link in self.metadata["footer_links"]:
            # there might be a link to Main-Page which would try (and fail)
            # to create wikiHo Main-Page (already added a redirect in homepage)
            if link.path and link.path != "Main-Page":
                self.scrape_article(link.path, remove_all_links=True)

    def scrape_related_articles(self):
        """Scrape and create all pages found in section related links"""
        for link in list(self.related_articles):
            if self.scrape_article(link) is None:
                self.related_articles.remove(link)

    def scrape_categories(self):
        logger.info("Starting scraping from categories")

        recurse = True  # will fetch sub-categories
        for category in self.conf.categories:
            if category.endswith("/"):
                category = category[:-1]
                recurse = False
            self.scrape_category(category, recurse)

    def scrape_category(self, category: str, recurse: bool = True):
        if category in self.categories or category in self.rewriter.excluded_categories:
            return
        self.categories.add(category)

        logger.info(f"> Category:{category} ({recurse=})")

        nb_pages, sub_categories = self.scrape_category_page(
            category, page_num=1, recurse=recurse
        )

        if nb_pages > 1:
            for page_num in range(2, nb_pages + 1):
                self.scrape_category_page(category, page_num=page_num, recurse=False)

        for sub_category in sub_categories or []:
            self.scrape_category(sub_category)

    def scrape_category_page(self, category: str, page_num: int, recurse: bool):

        category_url = f"/{self.metadata['category_prefix']}:{category}"
        params = {}
        if page_num > 1:
            logger.info(f">> Category:{category} (page={page_num})")
            params = {"pg": page_num}

        try:
            soup, paths = get_soup(category_url, **params)
        except requests.exceptions.HTTPError as exc:
            # don't fail on missing Category (#46)
            if exc.response.status_code == 404:
                logger.warning(">>> HTTP 404, skipping.")
                self.missing_categories.add(category_url)
                return 0, []
            raise exc

        fix_pagination_links(soup)

        articles = set()
        for link in soup.select("#cat_all .responsive_thumb a"):
            articles.add(article_ident_for(link.attrs.get("href")))

        for article in articles:
            if not self.scrape_article(article):
                missing_url = to_url(f"/{article}")
                for a in soup.find_all("a", href=missing_url):
                    del a["href"]
                self.record_missing_url(missing_url)
            # break  # only one article per page

        nb_pages = len(soup.select("#large_pagination ul li"))

        sub_categories = get_subcategories_from(soup, recurse)

        # extract and clean main content
        content = soup.select("div#content_wrapper")[0]

        black_list = (
            "script",
            "noscript > img",
            "noscript:empty",
            "#cat_wca",
            ".cat-promo",
        )
        for selector in black_list:
            _ = [elem.decompose() for elem in content.select(selector)]

        # zim article path
        path = paths.pop()
        if page_num > 1:
            path += f"_pg={page_num}"
        # some categories include a `/`. ex: Système-Macintosh/Apple
        to_root = "./" + ("../" * path.count("/"))

        title = soup.find("title").string
        page = self.env.get_template("category.html").render(
            to_root=to_root,
            body_classes=" ".join(soup.find("body").attrs.get("class", [])),
            content=self.rewriter.rewrite(content.decode_contents(), to_root=to_root),
            page_linked_styles=self.get_style_urls(soup),
            viewport_classes=" ".join(
                soup.find(attrs={"id": "mw-mf-viewport"}).attrs.get("class", [])
                + ["wikihow-category"]
            ),
            footer_links=self.metadata["footer_links"],
            breadcrumbs=get_footer_crumbs_from(soup, self.rewriter.excluded_categories),
            title=title,
            **self.env_context,
        )
        with self.lock:
            self.creator.add_item_for(
                path=path,
                title=title,
                content=page,
                mimetype="text/html",
                is_front=True,
            )

        for redir_path in paths:
            with self.lock:
                self.creator.add_redirect(
                    path=redir_path + (f"_pg={page_num}" if page_num > 1 else ""),
                    target_path=path,
                )

        return nb_pages, sub_categories

    def scrape_article(self, article, remove_all_links=False):
        if article in self.articles or article in self.rewriter.excluded_articles:
            return
        self.articles.add(article)

        logger.info(f">> Article:{article}")

        try:
            soup, _ = get_soup(f"/{article}")
        except requests.exceptions.HTTPError as exc:
            if exc.response.status_code == 404:
                logger.warning(">>> HTTP 404, skipping.")
                return False
            raise exc

        # extract and clean main content
        title = soup.find("title").string
        content = soup.select("div#content_wrapper")[0]

        black_list = (
            "script",
            ".pdf_link",
            "#side_follow",
            "#sidebar_share",
            "#other_languages",
            "#article_rating_mobile",
            "#end_options",
            ".section.tips",
            ".section.bss_share_container",
            ".wh_ad_inner, .wh_ad_active, [data-service=adsense], .wh_ad_spacing",
            "noscript > img",
            "noscript:empty",
            # User success stories
            "#userreviews_mobile",
            # Q&A block
            "#qa >:not(#qa_answered_questions_container)",
            # Expert Q&A
            ".qa_heading sticking",
            "#qa_answered_questions_container #qa_aq_search_container, "
            "#qa_answered_questions_container #qa_add_curated_question",
            ".qa_answer_footer",
            ".qa_no_answered",
            "#qa_see_more_answered",
            "#userreview_recipe_cta",
            "#aboutthisarticle > div.page_stats.section_text",
        )
        for selector in black_list:
            _ = [elem.decompose() for elem in content.select(selector)]

        # Remove link of page of author
        for link in content.select(
            "#coauthor_byline #byline_info > a[href], .sp_expert_text a[href]"
        ):
            del link.attrs["href"]

        if remove_all_links:
            for elem in content.select("a[href]"):
                del elem.attrs["href"]

        for link in content.select("div.section.relatedwikihows a[href]"):
            rel_article = re.sub(r"^/", "", normalize_ident(link.attrs["href"]))
            if rel_article not in self.articles:
                self.related_articles.add(rel_article)

        self.handle_videos_for(soup)

        # some articles include a `/`. ex: Système-Macintosh/Apple
        to_root = "./" + ("../" * article.count("/"))
        page = self.env.get_template("article.html").render(
            to_root=to_root,
            body_classes=" ".join(soup.find("body").attrs.get("class", [])),
            content=self.rewriter.rewrite(content.decode_contents(), to_root=to_root),
            page_linked_styles=self.get_style_urls(soup),
            viewport_classes=" ".join(
                soup.find(attrs={"id": "mw-mf-viewport"}).attrs.get("class", [])
                + ["wikihow-article"]
            ),
            footer_links=self.metadata["footer_links"],
            breadcrumbs=get_footer_crumbs_from(soup, self.rewriter.excluded_categories),
            title=title,
            **self.env_context,
        )
        with self.lock:
            self.creator.add_item_for(
                path=article,
                title=title,
                content=page,
                mimetype="text/html",
                is_front=True,
            )
        return True

    def record_missing_url(self, url):
        self.missing_articles.add(url)

        if len(self.missing_articles) >= MAX_HTTP_404_THRESHOLD:
            logger.debug(
                f"Maximum HTTP 404 threshold reached ({MAX_HTTP_404_THRESHOLD})"
            )

    def handle_videos_for(self, soup: bs4.element.Tag):
        # youtube video blocks
        if self.conf.without_videos:
            # remove link to video block in TOC
            # note that #summaryvideo_toc is different and is not Youtube
            _ = [elem.decompose() for elem in soup.select("#othervideo_toc")]
            # English has a .section.video selector for the video block we target
            # but the class name is localized so it differs from languages.
            # we thus remove the #video parent instead
            _ = [elem.parent.decompose() for elem in soup.select("#video")]
        else:
            for iframe in soup.select(".embedvideocontainer iframe.embedvideo"):
                path = Global.vidgrabber.defer(url=iframe.get("data-src"))
                if path is None:
                    iframe.decompose()
                    continue

                poster = Global.imager.defer(
                    Global.vidgrabber.youtube_poster_url(iframe.get("data-src"))
                )
                iframe.replace_with(
                    get_soup_of(
                        self.env.get_template("video.html").render(
                            path=path,
                            poster=poster,
                            classes=["youtube"],
                            video_format=self.conf.video_format,
                            width=iframe.attrs.get("width", "728"),
                            height=iframe.attrs.get("height", "428"),
                            autoplay="autoplay" in iframe.attrs.get("allow", ""),
                            controls=True,
                            preload=False,
                        ),
                        unwrap=True,
                    )
                )

        # record the video summary section if present.
        # allows distinguishing between step videos and summary one
        try:
            video_summary = soup.select(".summary_with_video")[0]
        except IndexError:
            video_summary = None

        # main-content (step) video hosted by wikiHow
        for video in soup.select(".video-player .video-container video"):
            # skip our own videos
            if "wikihow2zim" in video.attrs.get("class"):
                continue

            url = video.attrs.get("src")
            if not url and not video.attrs.get("data-src"):
                # missing source
                video.decompose()
                continue
            if not url:
                url = to_url(f"/video{video.attrs.get('data-src')}")

            path = Global.vidgrabber.defer(url=to_url(url))
            if path is None:
                continue

            poster_path = Global.imager.defer(
                url=to_url(video.attrs.get("poster", video.attrs.get("data-poster")))
            )
            # remove extra “controls” and watermark (from .video-player) :requires JS
            for elem in video.parent.parent.select(".m-video-controls, .m-video-wm"):
                elem.decompose()

            show_controls = video_summary and video in video_summary.select("video")
            video.replace_with(
                get_soup_of(
                    self.env.get_template("video.html").render(
                        path=path,
                        poster=poster_path,
                        classes=" ".join(video.attrs.get("class")),
                        video_format=self.conf.video_format,
                        autoplay=True,
                        muted="muted" in video.attrs,
                        loop="loop" in video.attrs,
                        playsinline="playsinline" in video.attrs,
                        controls=show_controls,
                        preload=True,
                    )
                )
            )

        # related articles links using video as thumbnail
        for video in soup.select("#relatedwikihows .related-wh video"):
            if "wikihow2zim" in video.attrs.get("class"):
                continue

            url = video.attrs.get("src")
            if not url and not video.attrs.get("data-src"):
                # missing source
                video.decompose()
                continue

            if not url:
                url = to_url(f"/video{video.attrs.get('data-src')}")

            poster_path = Global.imager.defer(
                url=to_url(video.attrs.get("poster", video.attrs.get("data-poster")))
            )
            path = Global.vidgrabber.defer(url=to_url(url))
            if path is None:
                video.decompose()
                continue

            video.replace_with(
                get_soup_of(
                    self.env.get_template("video.html").render(
                        path=path,
                        poster=poster_path,
                        classes=" ".join(video.attrs.get("class")),
                        video_format=self.conf.video_format,
                        autoplay=True,
                        width=video.attrs.get("width", "342"),
                        height=video.attrs.get("height", "184"),
                        alt=video.attrs.get("alt"),
                        muted="muted" in video.attrs,
                        loop="loop" in video.attrs,
                        playsinline="playsinline" in video.attrs,
                        controls=False,
                        preload=True,
                    )
                )
            )

    def check_dom_integrity(self):
        """Checking wikiHow DOM Integrity

        Verifying that the elements we rely on for the scrape are in place
        so we can fail early if not: meaning source changed significantly"""

        logger.info("Ensuring source site DOM Integrity")

        logger.debug("> checking CategoryListing")

        soup, _ = get_soup("/Special:CategoryListing")
        if not soup.select("#content_wrapper"):
            raise DomIntegrityError("#content_wrapper not found")
        category_links = soup.select("#catlist_container #catlist a")

        if not category_links:
            raise DomIntegrityError("No links in #catlist_container")

        # selecting a randon element in the list of links.
        # these links should all be Category links
        # we'll reuse selected link to check Category page later
        category_link = category_links[
            random.randint(0, len(category_links) - 1)
        ].attrs.get("href")
        if not category_link:
            raise DomIntegrityError("Category link has no href")

        # Making sure it was an actual Category link
        category_id = cat_ident_for(category_link)
        if not re.findall(":", category_link):
            raise DomIntegrityError("has not category link")

        category_req_url = normalize_ident(to_url(f"/Category:{category_id}"))
        resp = requests.get(category_req_url)
        if resp.status_code != 200:
            raise DomIntegrityError(f"Category link if not valid ({resp})")

        logger.debug("> checking Category Page")

        # Category page is mostly a grid listing articles in the Category
        soup, _ = get_soup(f"/Category:{category_id}")
        if not soup.select("#cat_all > div.cat_grid"):
            raise DomIntegrityError("Article list not found in #cat_grid")

        logger.debug("> checking Article Page")

        # Using Randomizer to select a random article from source website
        soup, _ = get_soup("/Special:Randomizer")

        # Checking that we have a title where expected
        if not soup.select("#content_inner > div.pre-content h1"):
            raise DomIntegrityError("Article title not found (h1)")

    def run(self):
        s3_storage = (
            setup_s3_and_check_credentials(self.conf.s3_url_with_credentials)
            if self.conf.s3_url_with_credentials
            else None
        )
        s3_msg = (
            f"\n"
            f"  using cache: {s3_storage.url.netloc} "
            f"with bucket: {s3_storage.bucket_name}"
            if s3_storage
            else ""
        )
        del s3_storage

        logger.info(
            f"Starting scraper with:\n"
            f"  language: {self.conf.language['english']}"
            f" ({self.conf.domain})\n"
            f"  output_dir: {self.conf.output_dir}\n"
            f"  build_dir: {self.build_dir}\n"
            f"  categories: "
            f"{', '.join(self.conf.categories)if self.conf.categories else 'all'}"
            f"{s3_msg}"
        )

        if not self.conf.skip_dom_check:
            self.check_dom_integrity()

        Global.metadata = self.get_online_metadata()
        logger.debug(
            f"homepage_name: {self.metadata['homepage_name']}\n"
            f"category_prefix: {self.metadata['category_prefix']}\n"
            f"url_special_category: {self.metadata['url_special_category']}\n"
            f"dir: {self.metadata['dir']}\n"
            f"title: {self.metadata['title']}\n"
            f"description: {self.metadata['description']}\n"
            f"icon: {self.metadata['icon']}\n"
            f"favicon: {self.metadata['favicon']}\n"
            f"logo: {self.metadata['logo']}"
        )
        self.sanitize_inputs()

        logger.debug("Starting Zim creation")
        Global.setup()
        self.creator.start()

        try:
            self.add_illustrations()
            self.add_assets()
            self.env_context.update(
                {
                    "dir": self.metadata["dir"],
                    "lang": self.conf.lang_code,
                    "linked_styles": self.metadata["linked_styles"],
                    "inline_digest": self.metadata["inline_digest"],
                    "homepage_name": self.metadata["homepage_name"],
                }
            )

            self.build_exclude_lists()

            if not self.conf.categories:
                self.build_categories_list()

            # start adding ZIM pages
            self.add_homepage()

            if not self.conf.skip_footer_links:
                self.scrape_footer_articles()

            if self.conf.single_article:
                self.scrape_article(self.conf.single_article)
            else:
                self.scrape_categories()

            if not self.conf.skip_relateds:
                self.scrape_related_articles()

            logger.info(
                f"Stats: {len(self.categories)} categories, "
                f"{len(self.articles)} articles, "
                f"{len(self.missing_categories)} missing categories, "
                f"{len(self.missing_articles)} missing articles, "
                f"{len(self.related_articles)} related articles, "
                f"{self.imager.nb_requested} images, "
                f"{self.vidgrabber.nb_requested} videos"
            )

            logger.info("Awaiting images")
            Global.img_executor.shutdown()

            logger.info("Awaiting videos")
            Global.video_executor.shutdown()

        except Exception as exc:
            # request Creator not to create a ZIM file on finish
            self.creator.can_finish = False
            if isinstance(exc, KeyboardInterrupt):
                logger.error("KeyboardInterrupt, exiting.")
            else:
                logger.error(f"Interrupting process due to error: {exc}")
                logger.exception(exc)
            self.imager.abort()
            Global.img_executor.shutdown(wait=False)
            return 1
        else:
            logger.info("Finishing ZIM file")
            # we need to release libzim's resources.
            # currently does nothing but crash if can_finish=False but that's awaiting
            # impl. at libkiwix level
            with self.lock:
                self.creator.finish()
            logger.info(
                f"Finished Zim {self.creator.filename.name} "
                f"in {self.creator.filename.parent}"
            )
        finally:
            self.cleanup()



# -*- coding: utf-8 -*-

import pathlib
import tempfile
import urllib.parse
from dataclasses import dataclass, field
from typing import List, Optional

from zimscraperlib.i18n import get_language_details

ROOT_DIR = pathlib.Path(__file__).parent
NAME = ROOT_DIR.name
DEFAULT_HOMEPAGE = "Main-Page"
MAX_HTTP_404_THRESHOLD = 200

with open(ROOT_DIR.joinpath("VERSION"), "r") as fh:
    VERSION = fh.read().strip()

SCRAPER = f"{NAME} {VERSION}"
IMAGES_ENCODER_VERSION = 1
VIDEOS_ENCODER_VERSION = 1
URLS = {
    "en": "https://www.wikihow.com",
    "ar": "https://ar.wikihow.com",
    "cs": "https://www.wikihow.cz",
    "de": "https://de.wikihow.com",
    "es": "https://es.wikihow.com",
    "fa": "https://www.wikihowfarsi.com",
    "fr": "https://fr.wikihow.com",
    "hi": "https://hi.wikihow.com",
    "id": "https://id.wikihow.com",
    "it": "https://www.wikihow.it",
    "ja": "https://www.wikihow.jp",
    "ko": "https://ko.wikihow.com",
    "nl": "https://nl.wikihow.com",
    "pt": "https://pt.wikihow.com",
    "ru": "https://ru.wikihow.com",
    "th": "https://th.wikihow.com",
    "tr": "https://www.wikihow.com.tr",
    "vi": "https://www.wikihow.vn",
    "zh": "https://zh.wikihow.com",
}


@dataclass
class Conf:
    required = [
        "lang_code",
        "output_dir",
    ]

    lang_code: str = ""
    language: dict = field(default_factory=dict)
    main_url: str = ""

    # zim params
    name: str = ""
    title: Optional[str] = ""
    description: Optional[str] = ""
    author: Optional[str] = ""
    publisher: Optional[str] = ""
    fname: Optional[str] = ""
    tag: List[str] = field(default_factory=list)

    # customization
    icon: Optional[str] = ""

    # filesystem
    _output_dir: Optional[str] = "."
    _tmp_dir: Optional[str] = "."
    output_dir: Optional[pathlib.Path] = None
    tmp_dir: Optional[pathlib.Path] = None

    # performances
    nb_threads: Optional[int] = -1
    s3_url_with_credentials: Optional[str] = ""

    # quality
    without_videos: Optional[bool] = False
    without_external_links: Optional[bool] = False
    exclude: Optional[str] = ""
    low_quality: Optional[bool] = False
    video_format: Optional[str] = "webm"

    # debug/devel
    categories: List[str] = field(default_factory=list)
    build_dir_is_tmp_dir: Optional[bool] = False
    keep_build_dir: Optional[bool] = False
    debug: Optional[bool] = False
    delay: Optional[float] = 0
    stats_filename: Optional[str] = None
    skip_dom_check: Optional[bool] = False
    skip_footer_links: Optional[bool] = False
    skip_relateds: Optional[bool] = False
    single_article: Optional[str] = ""

    @staticmethod
    def get_url(lang_code):
        return urllib.parse.urlparse(URLS[lang_code])

    @property
    def domain(self):
        return self.main_url.netloc

    @property
    def s3_url(self):
        return self.s3_url_with_credentials

    def __post_init__(self):
        self.main_url = Conf.get_url(self.lang_code)
        self.language = get_language_details(self.lang_code)
        self.output_dir = pathlib.Path(self._output_dir).expanduser().resolve()
        self.output_dir.mkdir(parents=True, exist_ok=True)

        self.tmp_dir = pathlib.Path(self._tmp_dir).expanduser().resolve()
        self.tmp_dir.mkdir(parents=True, exist_ok=True)
        if self.build_dir_is_tmp_dir:
            self.build_dir = self.tmp_dir
        else:
            self.build_dir = pathlib.Path(
                tempfile.mkdtemp(prefix=f"{self.domain}_", dir=self.tmp_dir)
            )
        self.build_dir.joinpath("videos").mkdir(parents=True, exist_ok=True)

        if self.stats_filename:
            self.stats_filename = pathlib.Path(self.stats_filename).expanduser()
            self.stats_filename.parent.mkdir(parents=True, exist_ok=True)

        # support semi-colon separated tags as well
        if self.tag:
            for tag in self.tag.copy():
                if ";" in tag:
                    self.tag += [p.strip() for p in tag.split(";")]
                    self.tag.remove(tag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import queue
import threading
from typing import Callable

from .shared import logger

_shutdown = False
# Lock that ensures that new workers are not created while the interpreter is
# shutting down. Must be held while mutating _threads_queues and _shutdown.
_global_shutdown_lock = threading.Lock()


def excepthook(args):
    logger.error(f"UNHANDLED Exception in {args.thread.name}: {args.exc_type}")
    logger.exception(args.exc_value)


threading.excepthook = excepthook


class Executor(queue.Queue):
    """Custom FIFO queue based Executor that's less generic than ThreadPoolExec one

    Providing more flexibility for the use cases we're interested about:
    - halt immediately (sort of) upon exception (if requested)
    - able to join() then restart later to accomodate successive steps

    See: https://github.com/python/cpython/blob/3.8/Lib/concurrent/futures/thread.py
    """

    def __init__(self, queue_size: int = 10, nb_workers: int = 1, prefix: str = "T-"):
        super().__init__(queue_size)
        self.prefix = prefix
        self._shutdown_lock = threading.Lock()
        self.nb_workers = nb_workers
        self.exceptions = []

    @property
    def exception(self):
        """Exception raises in any thread, if any"""
        try:
            return self.exceptions[0:1].pop()
        except IndexError:
            return None

    @property
    def alive(self):
        """whether it should continue running"""
        return not self._shutdown

    def submit(self, task: Callable, **kwargs):
        """Submit a callable and its kwargs for execution in one of the workers"""
        with self._shutdown_lock, _global_shutdown_lock:
            if not self.alive:
                raise RuntimeError("cannot submit task to dead executor")
            if _shutdown:
                raise RuntimeError("cannot submit task after " "interpreter shutdown")

        while True:
            try:
                self.put((task, kwargs), block=True, timeout=3.0)
            except queue.Full:
                if self.no_more:
                    logger.debug("rejecting task: queue full and currently `join`ing")
                    break
            else:
                break

    def start(self):
        """Enable executor, starting requested amount of workers

        Workers are started always, not provisioned dynamicaly"""
        self.drain()
        self.release_halt()
        self._workers = set()
        self._shutdown = False
        self.exceptions[:] = []

        for n in range(self.nb_workers):
            t = threading.Thread(target=self.worker, name=f"{self.prefix}{n}")
            t.daemon = True
            t.start()
            self._workers.add(t)

    def worker(self):
        while self.alive or self.no_more:
            try:
                func, kwargs = self.get(block=True, timeout=2.0)
            except queue.Empty:
                if self.no_more:
                    break
                continue
            except TypeError:
                # received None from the queue. most likely shuting down
                return

            raises = kwargs.pop("raises") if "raises" in kwargs.keys() else False
            callback = kwargs.pop("callback") if "callback" in kwargs.keys() else None
            dont_release = kwargs.pop("dont_release", False)

            try:
                func(**kwargs)
            except Exception as exc:
                logger.error(f"Error processing {func} with {kwargs=}")
                logger.exception(exc)
                if raises:
                    self.exceptions.append(exc)
                    self.shutdown()
            finally:
                # user will manually release the queue for this task.
                # most likely in a libzim-written callback
                if not dont_release:
                    self.task_done()
                if callback:
                    callback.__call__()

    def drain(self):
        """Empty the queue without processing the tasks (tasks will be lost)"""
        while True:
            try:
                self.get_nowait()
            except queue.Empty:
                break

    def join(self):
        """Await completion of workers, requesting them to stop taking new task"""
        logger.debug(f"joining all threads for {self.prefix}")
        self.no_more = True
        for t in self._workers:
            e = threading.Event()
            while t.is_alive():
                t.join(1)
                e.wait(timeout=2)
        logger.debug(f"all threads joined for {self.prefix}")

    def release_halt(self):
        """release the `no_more` flag preventing workers from taking up tasks"""
        self.no_more = False

    def shutdown(self, wait=True):
        """stop the executor, either somewhat immediately or awaiting completion"""
        logger.debug(f"shutting down executor {self.prefix} with {wait=}")
        with self._shutdown_lock:
            self._shutdown = True

            # Drain all work items from the queue
            if not wait:
                self.drain()
        if wait:
            self.join()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import io
import pathlib
import re
import urllib.parse
from typing import Optional

from kiwixstorage import KiwixStorage, NotFoundError
from PIL import Image
from zimscraperlib.download import stream_file
from zimscraperlib.image.optimization import optimize_webp

from .constants import IMAGES_ENCODER_VERSION
from .shared import Global
from .utils import get_digest, get_version_ident_for, normalize_ident, to_url

logger = Global.logger


class Imager:
    def __init__(self):
        self.aborted = False
        # list of source URLs that we've processed and added to ZIM
        self.handled = set()
        self.nb_requested = 0
        self.nb_done = 0

        Global.img_executor.start()

    def abort(self):
        """request imager to cancel processing of futures"""
        self.aborted = True

    def get_image_data(self, url: str) -> io.BytesIO:
        """Bytes stream of an optimized version of source image

        Bitmap images are converted to WebP and optimized
        SVG images are kept as is"""
        src, webp = io.BytesIO(), io.BytesIO()
        stream_file(url=url, byte_stream=src)

        if pathlib.Path(url).suffix == ".svg" or "/math/render/svg/" in url:
            return src

        with Image.open(src) as img:
            img.save(webp, format="WEBP")
        del src
        return optimize_webp(
            src=webp,
            lossless=False,
            quality=60,
            method=6,
        )

    def get_s3_key_for(self, url: str) -> str:
        """S3 key to use for that url"""
        return re.sub(r"^(https?)://", r"\1/", url)

    def get_path_for(self, url: urllib.parse.ParseResult) -> str:
        suffix = ".svg" if url.path.endswith(".svg") else ".webp"
        digest = get_digest(url.geturl())
        return f"images/{digest}-{normalize_ident(pathlib.Path(url.path).stem)}{suffix}"

    def defer(
        self,
        url: str,
        path: Optional[str] = None,
    ) -> str:
        """request full processing of url, returning in-zim path immediately"""

        # find actual URL should it be from a provider
        try:
            url = urllib.parse.urlparse(to_url(url))
        except Exception:
            logger.warning(f"Can't parse image URL `{url}`. Skipping")
            return

        if url.scheme not in ("http", "https"):
            logger.warning(f"Not supporting image URL `{url.geturl()}`. Skipping")
            return

        # skip processing if we already processed it or have it in pipe
        digest = get_digest(url.geturl())
        path = self.get_path_for(url) if path is None else path

        if digest in self.handled:
            logger.debug(f"URL `{url.geturl()}` already processed.")
            return path

        # record that we are processing this one
        self.handled.add(digest)
        self.nb_requested += 1

        Global.img_executor.submit(
            self.process_image,
            url=url,
            path=path,
            mimetype="image/svg+xml" if path.endswith(".svg") else "image/webp",
            dont_release=True,
        )

        return path

    def once_done(self):
        """default callback for single image processing"""
        self.nb_done += 1
        logger.debug(f"Images {self.nb_done}/{self.nb_requested}")

    def process_image(self, url: str, path: str, mimetype: str) -> str:
        """download image from url or S3 and add to Zim at path. Upload if req."""

        if self.aborted:
            return

        # just download, optimize and add to ZIM if not using S3
        if not Global.conf.s3_url:
            with Global.lock:
                Global.creator.add_item_for(
                    path=path,
                    content=self.get_image_data(url.geturl()).getvalue(),
                    mimetype=mimetype,
                    callback=self.once_done,
                )
            return path

        # we are using S3 cache
        ident = get_version_ident_for(url.geturl())
        if ident is None:
            logger.error(f"Unable to query {url.geturl()}. Skipping")
            return path

        key = self.get_s3_key_for(url.geturl())
        s3_storage = KiwixStorage(Global.conf.s3_url)
        meta = {"ident": ident, "encoder_version": str(IMAGES_ENCODER_VERSION)}

        download_failed = False  # useful to trigger reupload or not
        try:
            logger.debug(f"Attempting download of S3::{key} into ZIM::{path}")
            fileobj = io.BytesIO()
            s3_storage.download_matching_fileobj(key, fileobj, meta=meta)
        except NotFoundError:
            # don't have it, not a donwload error. we'll upload after processing
            pass
        except Exception as exc:
            logger.error(f"failed to download {key} from cache: {exc}")
            logger.exception(exc)
            download_failed = True
        else:
            with Global.lock:
                Global.creator.add_item_for(
                    path=path,
                    content=fileobj.getvalue(),
                    mimetype=mimetype,
                    callback=self.once_done,
                )
            return path

        # we're using S3 but don't have it or failed to download
        try:
            fileobj = self.get_image_data(url.geturl())
        except Exception as exc:
            logger.error(f"Failed to download/convert/optim source  at {url.geturl()}")
            logger.exception(exc)
            return path

        with Global.lock:
            Global.creator.add_item_for(
                path=path,
                content=fileobj.getvalue(),
                mimetype=mimetype,
                callback=self.once_done,
            )

        # only upload it if we didn't have it in cache
        if not download_failed:
            logger.debug(f"Uploading {url.geturl()} to S3::{key} with {meta}")
            try:
                s3_storage.upload_fileobj(fileobj=fileobj, key=key, meta=meta)
            except Exception as exc:
                logger.error(f"{key} failed to upload to cache: {exc}")

        return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
# pylint: disable=cyclic-import

import datetime
import logging
import threading

from zimscraperlib.logging import getLogger as lib_getLogger
from zimscraperlib.zim.creator import Creator

from .constants import DEFAULT_HOMEPAGE, NAME


class Global:
    """Shared context accross all scraper components"""

    debug = False
    logger = lib_getLogger(
        NAME,
        level=logging.INFO,
        log_format="[%(threadName)s::%(asctime)s] %(levelname)s:%(message)s",
    )
    conf = None

    metadata = {}

    creator = None
    imager = None
    rewriter = None
    lock = threading.Lock()

    @staticmethod
    def set_debug(value):
        Global.debug = value
        level = logging.DEBUG if value else logging.INFO
        Global.logger.setLevel(level)
        for handler in Global.logger.handlers:
            handler.setLevel(level)

    @staticmethod
    def setup():
        # order matters are there are references between them

        # images handled on a different queue.
        # mostly network I/O to retrieve and/or upload image.
        # if not in S3 bucket, convert/optimize webp image
        # svg images, stored but not optimized
        from .executor import Executor

        Global.img_executor = Executor(
            queue_size=20,
            nb_workers=10,
            prefix="IMG-T-",
        )

        # without_videos means without Youtube videos but there are still plenty
        # of regular videos for animations.
        # We use a single video worker for videos if Youtube is enabled to prevent
        # blacklisting of the host IP by Youtube
        # this should be smarter in processing non-youtube videos in parallel and
        # limit youtube ones to a single worker in every cases
        Global.video_executor = Executor(
            queue_size=20,
            nb_workers=10 if Global.conf.without_videos else 1,
            prefix="VID-T-",
        )

        from .imager import Imager

        Global.imager = Imager()

        from .videos import VideoGrabber

        Global.vidgrabber = VideoGrabber()

        from .rewriter import Rewriter

        Global.rewriter = Rewriter()

        Global.creator = Creator(
            filename=Global.conf.output_dir.joinpath(Global.conf.fname),
            main_path=DEFAULT_HOMEPAGE,
            favicon_path="illustration",
            language=Global.conf.language["iso-639-3"],
            title=Global.conf.title,
            description=Global.conf.description,
            creator=Global.conf.author,
            publisher=Global.conf.publisher,
            name=Global.conf.name,
            tags=";".join(Global.conf.tags),
            date=datetime.date.today(),
        ).config_verbose(True)


class GlobalMixin:
    @property
    def conf(self):
        return Global.conf

    @property
    def metadata(self):
        return Global.metadata

    @property
    def creator(self):
        return Global.creator

    @property
    def lock(self):
        return Global.lock

    @property
    def imager(self):
        return Global.imager

    @property
    def vidgrabber(self):
        return Global.vidgrabber

    @property
    def executor(self):
        return Global.executor

    @property
    def rewriter(self):
        return Global.rewriter


logger = Global.logger



# -*- coding: utf-8 -*-

import pathlib
import subprocess

from setuptools import setup

root_dir = pathlib.Path(__file__).parent


def read(*names, **kwargs):
    with open(root_dir.joinpath(*names), "r") as fh:
        return fh.read()


print("Downloading JS dependencies...")
subprocess.run([str(root_dir.joinpath("get_js_deps.sh").resolve())], check=True)


setup(
    name="wikihow2zim",
    version=read("wikihow2zim", "VERSION").strip(),
    description="Make ZIM file from WikiHow articles",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Kiwix Team",
    author_email="dev@kiwix.org",
    url="https://kiwix.org/",
    keywords="kiwix zim offline wikihow",
    license="GPLv3+",
    packages=["wikihow2zim"],
    install_requires=[
        line.strip()
        for line in read("requirements.txt").splitlines()
        if not line.strip().startswith("#")
    ],
    zip_safe=False,
    include_package_data=True,
    package_data={"": ["VERSION", "templates/*", "assets/*"]},
    entry_points={
        "console_scripts": [
            "wikihow2zim=wikihow2zim.__main__:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
    python_requires=">=3.8",
)
